@@ -268,3 +268,15 @@ class CreateAgentSessionEvent(BaseEvent):
 				'allowed_domains': agent.browser_profile.allowed_domains if agent.browser_profile else [],
 			},
 		)
+
+
+class UpdateAgentSessionEvent(BaseEvent):
+	"""Event to update an existing agent session"""
+
+	# Model fields
+	id: str  # Session ID to update
+	user_id: str = Field(max_length=255)
+	device_id: str | None = Field(None, max_length=255)
+	browser_session_stopped: bool | None = None
+	browser_session_stopped_at: datetime | None = None
+	end_reason: str | None = Field(None, max_length=100)  # Why the session ended
