@@ -27,6 +27,8 @@ from browser_use.dom.views import (
 if TYPE_CHECKING:
 	from browser_use.browser.session import BrowserSession
 
+# Note: iframe limits are now configurable via BrowserProfile.max_iframes and BrowserProfile.max_iframe_depth
+
 
 class DomService:
 	"""
@@ -40,11 +42,20 @@ class DomService:
 	logger: logging.Logger
 
 	def __init__(
-		self, browser_session: 'BrowserSession', logger: logging.Logger | None = None, cross_origin_iframes: bool = False
+		self,
+		browser_session: 'BrowserSession',
+		logger: logging.Logger | None = None,
+		cross_origin_iframes: bool = False,
+		paint_order_filtering: bool = True,
+		max_iframes: int = 100,
+		max_iframe_depth: int = 5,
 	):
 		self.browser_session = browser_session
 		self.logger = logger or browser_session.logger
 		self.cross_origin_iframes = cross_origin_iframes
+		self.paint_order_filtering = paint_order_filtering
+		self.max_iframes = max_iframes
+		self.max_iframe_depth = max_iframe_depth
 
 	async def __aenter__(self):
 		return self
@@ -407,16 +418,24 @@ class DomService:
 		end = time.time()
 		cdp_timing = {'cdp_calls_total': end - start}
 
-		# DEBUG: Log snapshot info
+		# DEBUG: Log snapshot info and limit documents to prevent explosion
 		if snapshot and 'documents' in snapshot:
+			original_doc_count = len(snapshot['documents'])
+			# Limit to max_iframes documents to prevent iframe explosion
+			if original_doc_count > self.max_iframes:
+				self.logger.warning(
+					f'⚠️ Limiting processing of {original_doc_count} iframes on page to only first {self.max_iframes} to prevent crashes!'
+				)
+				snapshot['documents'] = snapshot['documents'][: self.max_iframes]
+
 			total_nodes = sum(len(doc.get('nodes', [])) for doc in snapshot['documents'])
-			self.logger.debug(
-				f'🔍 DEBUG: Snapshot contains {len(snapshot["documents"])} documents with {total_nodes} total nodes'
-			)
+			self.logger.debug(f'🔍 DEBUG: Snapshot contains {len(snapshot["documents"])} frames with {total_nodes} total nodes')
 			# Log iframe-specific info
 			for doc_idx, doc in enumerate(snapshot['documents']):
 				if doc_idx > 0:  # Not the main document
-					self.logger.debug(f'🔍 DEBUG: Document {doc_idx} has {len(doc.get("nodes", []))} nodes')
+					self.logger.debug(
+						f'🔍 DEBUG: Iframe #{doc_idx} {doc.get("frameId", "no-frame-id")} {doc.get("url", "no-url")} has {len(doc.get("nodes", []))} nodes'
+					)
 
 		return TargetAllTrees(
 			snapshot=snapshot,
@@ -431,6 +450,7 @@ class DomService:
 		target_id: TargetID,
 		initial_html_frames: list[EnhancedDOMTreeNode] | None = None,
 		initial_total_frame_offset: DOMRect | None = None,
+		iframe_depth: int = 0,
 	) -> EnhancedDOMTreeNode:
 		"""Get the DOM tree for a specific target.
 
@@ -438,6 +458,7 @@ class DomService:
 			target_id: Target ID of the page to get the DOM tree for.
 			initial_html_frames: List of HTML frame nodes encountered so far
 			initial_total_frame_offset: Accumulated coordinate offset
+			iframe_depth: Current depth of iframe nesting to prevent infinite recursion
 		"""
 
 		trees = await self._get_all_trees(target_id)
@@ -621,33 +642,67 @@ class DomService:
 				# TODO: hacky way to disable cross origin iframes for now
 				self.cross_origin_iframes and node['nodeName'].upper() == 'IFRAME' and node.get('contentDocument', None) is None
 			):  # None meaning there is no content
-				# Use get_all_frames to find the iframe's target
-				frame_id = node.get('frameId', None)
-				if frame_id:
-					all_frames, _ = await self.browser_session.get_all_frames()
-					frame_info = all_frames.get(frame_id)
-					iframe_document_target = None
-					if frame_info and frame_info.get('frameTargetId'):
-						# Get the target info for this iframe
-						targets = await self.browser_session.cdp_client.send.Target.getTargets()
-						iframe_document_target = next(
-							(t for t in targets['targetInfos'] if t['targetId'] == frame_info['frameTargetId']), None
-						)
-				else:
-					iframe_document_target = None
-				# if target actually exists in one of the frames, just recursively build the dom tree for it
-				if iframe_document_target:
-					self.logger.debug(f'Getting content document for iframe {node.get("frameId", None)}')
-					content_document = await self.get_dom_tree(
-						target_id=iframe_document_target.get('targetId'),
-						# TODO: experiment with this values -> not sure whether the whole cross origin iframe should be ALWAYS included as soon as some part of it is visible or not.
-						# Current config: if the cross origin iframe is AT ALL visible, then just include everything inside of it!
-						# initial_html_frames=updated_html_frames,
-						initial_total_frame_offset=total_frame_offset,
+				# Check iframe depth to prevent infinite recursion
+				if iframe_depth >= self.max_iframe_depth:
+					self.logger.debug(
+						f'Skipping iframe at depth {iframe_depth} to prevent infinite recursion (max depth: {self.max_iframe_depth})'
 					)
+				else:
+					# Check if iframe is visible and large enough (>= 200px in both dimensions)
+					should_process_iframe = False
 
-					dom_tree_node.content_document = content_document
-					dom_tree_node.content_document.parent_node = dom_tree_node
+					# First check if the iframe element itself is visible
+					if dom_tree_node.is_visible:
+						# Check iframe dimensions
+						if dom_tree_node.snapshot_node and dom_tree_node.snapshot_node.bounds:
+							bounds = dom_tree_node.snapshot_node.bounds
+							width = bounds.width
+							height = bounds.height
+
+							# Only process if iframe is at least 200px in both dimensions
+							if width >= 200 and height >= 200:
+								should_process_iframe = True
+								self.logger.debug(f'Processing cross-origin iframe: visible=True, width={width}, height={height}')
+							else:
+								self.logger.debug(
+									f'Skipping small cross-origin iframe: width={width}, height={height} (needs >= 200px)'
+								)
+						else:
+							self.logger.debug('Skipping cross-origin iframe: no bounds available')
+					else:
+						self.logger.debug('Skipping invisible cross-origin iframe')
+
+					if should_process_iframe:
+						# Use get_all_frames to find the iframe's target
+						frame_id = node.get('frameId', None)
+						if frame_id:
+							all_frames, _ = await self.browser_session.get_all_frames()
+							frame_info = all_frames.get(frame_id)
+							iframe_document_target = None
+							if frame_info and frame_info.get('frameTargetId'):
+								# Get the target info for this iframe
+								targets = await self.browser_session.cdp_client.send.Target.getTargets()
+								iframe_document_target = next(
+									(t for t in targets['targetInfos'] if t['targetId'] == frame_info['frameTargetId']), None
+								)
+						else:
+							iframe_document_target = None
+						# if target actually exists in one of the frames, just recursively build the dom tree for it
+						if iframe_document_target:
+							self.logger.debug(
+								f'Getting content document for iframe {node.get("frameId", None)} at depth {iframe_depth + 1}'
+							)
+							content_document = await self.get_dom_tree(
+								target_id=iframe_document_target.get('targetId'),
+								# TODO: experiment with this values -> not sure whether the whole cross origin iframe should be ALWAYS included as soon as some part of it is visible or not.
+								# Current config: if the cross origin iframe is AT ALL visible, then just include everything inside of it!
+								# initial_html_frames=updated_html_frames,
+								initial_total_frame_offset=total_frame_offset,
+								iframe_depth=iframe_depth + 1,
+							)
+
+							dom_tree_node.content_document = content_document
+							dom_tree_node.content_document.parent_node = dom_tree_node
 
 			return dom_tree_node
 
@@ -670,7 +725,7 @@ class DomService:
 
 		start = time.time()
 		serialized_dom_state, serializer_timing = DOMTreeSerializer(
-			enhanced_dom_tree, previous_cached_state
+			enhanced_dom_tree, previous_cached_state, paint_order_filtering=self.paint_order_filtering
 		).serialize_accessible_elements()
 
 		end = time.time()
