@@ -19,6 +19,8 @@ DEFAULT_INCLUDE_ATTRIBUTES = [
 	'title',
 	'type',
 	'checked',
+	# 'class',
+	'id',
 	'name',
 	'role',
 	'value',
@@ -51,6 +53,51 @@ DEFAULT_INCLUDE_ATTRIBUTES = [
 	'ax_name',
 ]
 
+STATIC_ATTRIBUTES = {
+	'class',
+	'id',
+	'name',
+	'type',
+	'placeholder',
+	'aria-label',
+	'title',
+	# 'aria-expanded',
+	'role',
+	'data-testid',
+	'data-test',
+	'data-cy',
+	'data-selenium',
+	'for',
+	'required',
+	'disabled',
+	'readonly',
+	'checked',
+	'selected',
+	'multiple',
+	'href',
+	'target',
+	'rel',
+	'aria-describedby',
+	'aria-labelledby',
+	'aria-controls',
+	'aria-owns',
+	'aria-live',
+	'aria-atomic',
+	'aria-busy',
+	'aria-disabled',
+	'aria-hidden',
+	'aria-pressed',
+	'aria-checked',
+	'aria-selected',
+	'tabindex',
+	'alt',
+	'src',
+	'lang',
+	'itemscope',
+	'itemtype',
+	'itemprop',
+}
+
 
 @dataclass
 class CurrentPageTargets:
@@ -90,7 +137,10 @@ class SimplifiedNode:
 	interactive_index: int | None = None
 
 	is_new: bool = False
+
+	ignored_by_paint_order: bool = False  # More info in dom/serializer/paint_order.py
 	excluded_by_parent: bool = False  # New field for bbox filtering
+	is_shadow_host: bool = False  # New field for shadow DOM hosts
 
 	def _clean_original_node_json(self, node_json: dict) -> dict:
 		"""Recursively remove children_nodes and shadow_roots from original_node JSON."""
@@ -113,6 +163,8 @@ class SimplifiedNode:
 		return {
 			'should_display': self.should_display,
 			'interactive_index': self.interactive_index,
+			'ignored_by_paint_order': self.ignored_by_paint_order,
+			'excluded_by_parent': self.excluded_by_parent,
 			'original_node': cleaned_original_node_json,
 			'children': [c.__json__() for c in self.children],
 		}
@@ -142,6 +194,17 @@ class DOMRect:
 	width: float
 	height: float
 
+	def to_dict(self) -> dict[str, Any]:
+		return {
+			'x': self.x,
+			'y': self.y,
+			'width': self.width,
+			'height': self.height,
+		}
+
+	def __json__(self) -> dict:
+		return self.to_dict()
+
 
 @dataclass(slots=True)
 class EnhancedAXProperty:
@@ -372,6 +435,7 @@ class EnhancedDOMTreeNode:
 			'node_type': self.node_type.name,
 			'node_name': self.node_name,
 			'node_value': self.node_value,
+			'is_visible': self.is_visible,
 			'attributes': self.attributes,
 			'is_scrollable': self.is_scrollable,
 			'session_id': self.session_id,
@@ -667,8 +731,9 @@ class EnhancedDOMTreeNode:
 		parent_branch_path = self._get_parent_branch_path()
 		parent_branch_path_string = '/'.join(parent_branch_path)
 
-		# Get attributes hash
-		attributes_string = ''.join(f'{key}={value}' for key, value in self.attributes.items())
+		attributes_string = ''.join(
+			f'{k}={v}' for k, v in sorted((k, v) for k, v in self.attributes.items() if k in STATIC_ATTRIBUTES)
+		)
 
 		# Combine both for final hash
 		combined_string = f'{parent_branch_path_string}|{attributes_string}'
@@ -753,11 +818,16 @@ class DOMInteractedElement:
 
 	def to_dict(self) -> dict[str, Any]:
 		return {
+			'node_id': self.node_id,
+			'backend_node_id': self.backend_node_id,
+			'frame_id': self.frame_id,
 			'node_type': self.node_type.value,
 			'node_value': self.node_value,
 			'node_name': self.node_name,
 			'attributes': self.attributes,
 			'x_path': self.x_path,
+			'element_hash': self.element_hash,
+			'bounds': self.bounds.to_dict() if self.bounds else None,
 		}
 
 	@classmethod
